"""
 Copyright 2024 Google LLC

 Licensed under the Apache License, Version 2.0 (the "License");
 you may not use this file except in compliance with the License.
 You may obtain a copy of the License at

      https://www.apache.org/licenses/LICENSE-2.0

 Unless required by applicable law or agreed to in writing, software
 distributed under the License is distributed on an "AS IS" BASIS,
 WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 See the License for the specific language governing permissions and
 limitations under the License.
 """

import logging

import dataflux_core
from google.cloud import storage


def _get_missing_permissions(storage_client: any, bucket_name: str,
                             project_name: str, required_perm: any):
    """Returns a list of missing permissions of the client from the required permissions list."""
    if not storage_client:
        storage_client = storage.Client(project=project_name)
    dataflux_core.user_agent.add_dataflux_user_agent(storage_client)
    bucket = storage_client.bucket(bucket_name)

    try:
        perm = bucket.test_iam_permissions(required_perm)
    except Exception as e:
        logging.exception(f"Error testing permissions: {e}")

    return [p for p in required_perm if p not in perm]



from typing import Any, Dict, Optional

import torch
from dataflux_core import user_agent
from google.cloud import storage
from lightning.pytorch.plugins.io import CheckpointIO


class DatafluxLightningCheckpoint(CheckpointIO):
    """A checkpoint manager for GCS using the :class:'CheckpointIO' interface"""

    def __init__(
        self,
        project_name: str,
        bucket_name: str,
        storage_client: Optional[storage.Client] = None,
    ):
        self.project_name = project_name
        self.bucket_name = bucket_name
        self.storage_client = storage_client
        if not storage_client:
            self.storage_client = storage.Client(project=self.project_name, )
        user_agent.add_dataflux_user_agent(self.storage_client)
        self.bucket = self.storage_client.bucket(self.bucket_name)

    def _parse_gcs_path(self, path: str) -> str:
        if not path or not (path.startswith("gcs://")
                            or path.startswith("gs://")):
            raise ValueError("Path needs to begin with gcs:// or gs://")
        path = path.split("//", maxsplit=1)
        if not path or len(path) < 2:
            raise ValueError("Bucket name must be non-empty")
        split = path[1].split("/", maxsplit=1)
        if len(split) == 1:
            bucket = split[0]
            prefix = ""
        else:
            bucket, prefix = split
        if not bucket:
            raise ValueError("Bucket name must be non-empty")
        if bucket != self.bucket_name:
            raise ValueError(
                f'Unexpected bucket name, expected {self.bucket_name} got {bucket}'
            )
        return prefix

    def save_checkpoint(
        self,
        checkpoint: Dict[str, Any],
        path: str,
        storage_options: Optional[Any] = None,
    ) -> None:
        key = self._parse_gcs_path(path)
        blob = self.bucket.blob(key)
        with blob.open("wb", ignore_flush=True) as blobwriter:
            torch.save(checkpoint, blobwriter)

    def load_checkpoint(
        self,
        path: str,
        map_location: Optional[Any] = None,
    ) -> Dict[str, Any]:
        key = self._parse_gcs_path(path)
        blob = self.bucket.blob(key)
        return torch.load(blob.open("rb"), map_location)

    def remove_checkpoint(
        self,
        path: str,
    ) -> None:
        key = self._parse_gcs_path(path)
        blob = self.bucket.blob(key)
        blob.delete()

    def teardown(self, ) -> None:
        pass



"""
 Copyright 2024 Google LLC

 Licensed under the Apache License, Version 2.0 (the "License");
 you may not use this file except in compliance with the License.
 You may obtain a copy of the License at

      https://www.apache.org/licenses/LICENSE-2.0

 Unless required by applicable law or agreed to in writing, software
 distributed under the License is distributed on an "AS IS" BASIS,
 WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 See the License for the specific language governing permissions and
 limitations under the License.
 """

from __future__ import annotations

import itertools
import logging
import math
import multiprocessing
import queue
import signal
import sys
import threading
import uuid
from typing import Iterator

from google.api_core.client_info import ClientInfo
from google.cloud import storage
from google.cloud.storage.retry import DEFAULT_RETRY

from dataflux_core import user_agent

# https://cloud.google.com/storage/docs/retry-strategy#python.
MODIFIED_RETRY = DEFAULT_RETRY.with_deadline(300.0).with_delay(initial=1.0,
                                                               multiplier=1.2,
                                                               maximum=45.0)

# https://cloud.google.com/storage/docs/composite-objects.
MAX_NUM_OBJECTS_TO_COMPOSE = 32

COMPOSED_PREFIX = "dataflux-composed-objects/"

current_composed_object = None


def compose(
    project_name: str,
    bucket_name: str,
    destination_blob_name: str,
    objects: list[tuple[str, int]],
    storage_client: object = None,
    retry_config: "google.api_core.retry.retry_unary.Retry" = MODIFIED_RETRY,
) -> object:
    """Compose the objects into a composite object, upload the composite object to the GCS bucket and returns it.

    Args:
        project_name: the name of the GCP project.
        bucket_name: the name of the GCS bucket that holds the objects to compose.
            The function uploads the the composed object to this bucket too.
        destination_blob_name: the name of the composite object to be created.
        objects: A list of tuples which indicate the object names and sizes (in bytes) in the bucket.
            Example: [("object_name_A", 1000), ("object_name_B", 2000)]
        storage_client: the google.cloud.storage.Client initialized with the project.
            If not defined, the function will initialize the client with the project_name.
        retry_config: The retry parameter to supply to the compose objects call.

    Returns:
        the "blob" of the composed object.
    """
    if len(objects) > MAX_NUM_OBJECTS_TO_COMPOSE:
        raise ValueError(
            f"{MAX_NUM_OBJECTS_TO_COMPOSE} objects allowed to compose, received {len(objects)} objects."
        )

    if storage_client is None:
        storage_client = storage.Client(project=project_name)
    user_agent.add_dataflux_user_agent(storage_client)

    bucket = storage_client.bucket(bucket_name)
    destination = bucket.blob(destination_blob_name)

    sources = list()
    for each_object in objects:
        blob_name = each_object[0]
        sources.append(bucket.blob(blob_name))

    destination.compose(sources, retry=retry_config)

    return destination


def decompose(
    project_name: str,
    bucket_name: str,
    composite_object_name: str,
    objects: list[tuple[str, int]],
    storage_client: object = None,
    retry_config: "google.api_core.retry.retry_unary.Retry" = MODIFIED_RETRY,
) -> list[bytes]:
    """Decompose the composite objects and return the decomposed objects contents in bytes.

    Args:
        project_name: the name of the GCP project.
        bucket_name: the name of the GCS bucket that holds the objects to compose.
            The function uploads the the composed object to this bucket too.
        composite_object_name: the name of the composite object to be created.
        objects: A list of tuples which indicate the object names and sizes (in bytes) in the bucket.
            Example: [("object_name_A", 1000), ("object_name_B", 2000)]
        storage_client: the google.cloud.storage.Client initialized with the project.
            If not defined, the function will initialize the client with the project_name.
        retry_config: The retry parameter supplied to the download_as_bytes call.

    Returns:
        the contents (in bytes) of the decomposed objects.
    """
    if storage_client is None:
        storage_client = storage.Client(project=project_name)
    user_agent.add_dataflux_user_agent(storage_client)

    res = []
    composed_object_content = download_single(
        storage_client,
        bucket_name,
        composite_object_name,
        retry_config=retry_config,
    )

    start = 0
    for each_object in objects:
        blob_size = each_object[1]
        content = composed_object_content[start:start + blob_size]
        res.append(content)
        start += blob_size

    if start != len(composed_object_content):
        logging.error(
            "decomposed object length = %s bytes, wanted = %s bytes.",
            start,
            len(composed_object_content),
        )
    return res


def download_single(
    storage_client: object,
    bucket_name: str,
    object_name: str,
    retry_config: "google.api_core.retry.retry_unary.Retry" = MODIFIED_RETRY,
) -> bytes:
    """Download the contents of this object as a bytes object and return it.

    Args:
        storage_client: the google.cloud.storage.Client initialized with the project.
        bucket_name: the name of the GCS bucket that holds the object.
        object_name: the name of the object to download.
        retry_config: The retry parameter supplied to the download_as_bytes call.

    Returns:
        the contents of the object in bytes.
    """
    bucket_handle = storage_client.bucket(bucket_name)
    blob = bucket_handle.blob(object_name)
    return blob.download_as_bytes(retry=retry_config)


class DataFluxDownloadOptimizationParams:
    """Parameters used to optimize DataFlux download performance.

    Attributes:
        max_composite_object_size: An integer indicating a cap for the maximum size of the composite object.

    """

    def __init__(self, max_composite_object_size):
        self.max_composite_object_size = max_composite_object_size


def df_download_thread(
    results_queue: queue.Queue[list[bytes]],
    project_name: str,
    bucket_name: str,
    objects: list[tuple[str, int]],
    storage_client: object = None,
    dataflux_download_optimization_params:
    DataFluxDownloadOptimizationParams = None,
    retry_config=MODIFIED_RETRY,
):
    """Threading helper that calls dataflux_download and places results onto queue.

    Args:
        results_queue: the queue on which to put all download results.
        project_name: the name of the GCP project.
        bucket_name: the name of the GCS bucket that holds the objects to compose.
            The function uploads the the composed object to this bucket too.
        objects: A list of tuples which indicate the object names and sizes (in bytes) in the bucket.
            Example: [("object_name_A", 1000), ("object_name_B", 2000)]
        storage_client: the google.cloud.storage.Client initialized with the project.
            If not defined, the function will initialize the client with the project_name.
        dataflux_download_optimization_params: the paramemters used to optimize the download performance.
        retry_config: The retry configuration to pass to all retryable download operations
    """
    result = dataflux_download(
        project_name,
        bucket_name,
        objects,
        storage_client,
        dataflux_download_optimization_params,
        # Always signify threading enabled so that signal handling is disabled.
        threading_enabled=True,
        retry_config=retry_config,
    )
    results_queue.put(result)


def dataflux_download_threaded(
    project_name: str,
    bucket_name: str,
    objects: list[tuple[str, int]],
    storage_client: object = None,
    dataflux_download_optimization_params:
    DataFluxDownloadOptimizationParams = None,
    threads: int = 1,
    retry_config=MODIFIED_RETRY,
) -> list[bytes]:
    """Perform the DataFlux download algorithm threaded to performantly download the object contents as bytes and return.

    Args:
        project_name: the name of the GCP project.
        bucket_name: the name of the GCS bucket that holds the objects to compose.
            The function uploads the the composed object to this bucket too.
        objects: A list of tuples which indicate the object names and sizes (in bytes) in the bucket.
            Example: [("object_name_A", 1000), ("object_name_B", 2000)]
        storage_client: the google.cloud.storage.Client initialized with the project.
            If not defined, the function will initialize the client with the project_name.
        dataflux_download_optimization_params: the paramemters used to optimize the download performance.
        threads: The number of threads on which to download at any given time.
        retry_config: The retry configuration to pass to all retryable download operations
    Returns:
        the contents of the object in bytes.
    """
    chunk_size = math.ceil(len(objects) / threads)
    chunks = []
    for i in range(threads):
        chunk = objects[i * chunk_size:(i + 1) * chunk_size]
        if chunk:
            chunks.append(chunk)
    results_queues = [queue.Queue() for _ in chunks]
    thread_list = []
    for i, chunk in enumerate(chunks):
        thread = threading.Thread(
            target=df_download_thread,
            args=(
                results_queues[i],
                project_name,
                bucket_name,
                chunk,
                storage_client,
                dataflux_download_optimization_params,
                retry_config,
            ),
        )
        thread_list.append(thread)
        thread.start()
    for thread in thread_list:
        thread.join()
    results = []
    for q in results_queues:
        while not q.empty():
            results.extend(q.get())
    return results


def dataflux_download_parallel(
    project_name: str,
    bucket_name: str,
    objects: list[tuple[str, int]],
    storage_client: object = None,
    dataflux_download_optimization_params:
    DataFluxDownloadOptimizationParams = None,
    parallelization: int = 1,
    retry_config=MODIFIED_RETRY,
) -> list[bytes]:
    """Perform the DataFlux download algorithm in parallel to download the object contents as bytes and return.

    Args:
        project_name: the name of the GCP project.
        bucket_name: the name of the GCS bucket that holds the objects to compose.
            The function uploads the the composed object to this bucket too.
        objects: A list of tuples which indicate the object names and sizes (in bytes) in the bucket.
            Example: [("object_name_A", 1000), ("object_name_B", 2000)]
        storage_client: the google.cloud.storage.Client initialized with the project.
            If not defined, the function will initialize the client with the project_name.
        dataflux_download_optimization_params: the paramemters used to optimize the download performance.
        parallelization: The number of parallel processes that will simultaneously execute the download.
        retry_config: The retry configuration to pass to all retryable download operations
    Returns:
        the contents of the object in bytes.
    """
    chunk_size = math.ceil(len(objects) / parallelization)
    chunks = []
    for i in range(parallelization):
        chunk = objects[i * chunk_size:(i + 1) * chunk_size]
        if chunk:
            chunks.append(chunk)
    with multiprocessing.Pool(processes=len(chunks)) as pool:
        results = pool.starmap(
            dataflux_download,
            ((
                project_name,
                bucket_name,
                chunk,
                storage_client,
                dataflux_download_optimization_params,
                False,
                retry_config,
            ) for chunk in chunks),
        )
        return list(itertools.chain.from_iterable(results))


def dataflux_download(
    project_name: str,
    bucket_name: str,
    objects: list[tuple[str, int]],
    storage_client: object = None,
    dataflux_download_optimization_params:
    DataFluxDownloadOptimizationParams = None,
    threading_enabled=False,
    retry_config=MODIFIED_RETRY,
) -> list[bytes]:
    """Perform the DataFlux download algorithm to download the object contents as bytes and return.

    Args:
        project_name: the name of the GCP project.
        bucket_name: the name of the GCS bucket that holds the objects to compose.
            The function uploads the the composed object to this bucket too.
        objects: A list of tuples which indicate the object names and sizes (in bytes) in the bucket.
            Example: [("object_name_A", 1000), ("object_name_B", 2000)]
        storage_client: the google.cloud.storage.Client initialized with the project.
            If not defined, the function will initialize the client with the project_name.
        dataflux_download_optimization_params: the paramemters used to optimize the download performance.
        retry_config: The retry configuration to pass to all retryable download operations
    Returns:
        the contents of the object in bytes.
    """
    if storage_client is None:
        storage_client = storage.Client(project=project_name)
    user_agent.add_dataflux_user_agent(storage_client)

    res = []
    max_composite_object_size = (
        dataflux_download_optimization_params.max_composite_object_size)

    i = 0
    # Register the cleanup signal handler for SIGINT.
    if not threading_enabled:
        signal.signal(signal.SIGINT, term_signal_handler)
    global current_composed_object
    while i < len(objects):
        curr_object_name = objects[i][0]
        curr_object_size = objects[i][1]

        if curr_object_size > max_composite_object_size:
            # Download the single object.
            curr_object_content = download_single(
                storage_client=storage_client,
                bucket_name=bucket_name,
                object_name=curr_object_name,
                retry_config=retry_config,
            )
            res.append(curr_object_content)
            i += 1
        else:
            # Dynamically compose and decompose based on the object size.
            objects_slice = []
            curr_size = 0

            while (i < len(objects) and curr_size <= max_composite_object_size
                   and len(objects_slice) < MAX_NUM_OBJECTS_TO_COMPOSE):
                curr_size += objects[i][1]
                objects_slice.append(objects[i])
                i += 1

            if len(objects_slice) == 1:
                object_name = objects_slice[0][0]
                curr_object_content = download_single(
                    storage_client=storage_client,
                    bucket_name=bucket_name,
                    object_name=object_name,
                    retry_config=retry_config,
                )
                res.append(curr_object_content)
            else:
                # If the number of objects > 1, we want to compose, download, decompose and delete the composite object.
                # Need to create a unique composite name to avoid mutation on the same object among processes.
                composed_object_name = COMPOSED_PREFIX + str(uuid.uuid4())
                composed_object = compose(
                    project_name,
                    bucket_name,
                    composed_object_name,
                    objects_slice,
                    storage_client,
                    retry_config=retry_config,
                )
                current_composed_object = composed_object
                res.extend(
                    decompose(
                        project_name,
                        bucket_name,
                        composed_object_name,
                        objects_slice,
                        storage_client,
                        retry_config=retry_config,
                    ))

                try:
                    composed_object.delete(retry=retry_config)
                    current_composed_object = None
                except Exception as e:
                    logging.exception(
                        f"exception while deleting the composite object: {e}")
    return res


def dataflux_download_lazy(
    project_name: str,
    bucket_name: str,
    objects: list[tuple[str, int]],
    storage_client: object = None,
    dataflux_download_optimization_params:
    DataFluxDownloadOptimizationParams = None,
    threading_enabled=False,
    retry_config: "google.api_core.retry.retry_unary.Retry" = MODIFIED_RETRY,
) -> Iterator[bytes]:
    """Perform the DataFlux download algorithm to download the object contents as bytes in a lazy fashion.

    Args:
        project_name: the name of the GCP project.
        bucket_name: the name of the GCS bucket that holds the objects to compose.
            The function uploads the the composed object to this bucket too.
        objects: A list of tuples which indicate the object names and sizes (in bytes) in the bucket.
            Example: [("object_name_A", 1000), ("object_name_B", 2000)]
        storage_client: the google.cloud.storage.Client initialized with the project.
            If not defined, the function will initialize the client with the project_name.
        dataflux_download_optimization_params: the paramemters used to optimize the download performance.
        retry_config: The retry parameter to supply to the compose objects call.
    Returns:
        An iterator of the contents of the object in bytes.
    """
    if storage_client is None:
        storage_client = storage.Client(project=project_name)
    user_agent.add_dataflux_user_agent(storage_client)

    max_composite_object_size = (
        dataflux_download_optimization_params.max_composite_object_size)

    i = 0
    # Register the cleanup signal handler for SIGINT.
    if not threading_enabled:
        signal.signal(signal.SIGINT, term_signal_handler)
    global current_composed_object
    while i < len(objects):
        curr_object_name = objects[i][0]
        curr_object_size = objects[i][1]

        if curr_object_size > max_composite_object_size:
            # Download the single object.
            curr_object_content = download_single(
                storage_client=storage_client,
                bucket_name=bucket_name,
                object_name=curr_object_name,
                retry_config=retry_config,
            )
            yield from [curr_object_content]
            i += 1
        else:
            # Dynamically compose and decompose based on the object size.
            objects_slice = []
            curr_size = 0

            while (i < len(objects) and curr_size <= max_composite_object_size
                   and len(objects_slice) < MAX_NUM_OBJECTS_TO_COMPOSE):
                curr_size += objects[i][1]
                objects_slice.append(objects[i])
                i += 1

            if len(objects_slice) == 1:
                object_name = objects_slice[0][0]
                curr_object_content = download_single(
                    storage_client=storage_client,
                    bucket_name=bucket_name,
                    object_name=object_name,
                    retry_config=retry_config,
                )
                yield from [curr_object_content]
            else:
                # If the number of objects > 1, we want to compose, download, decompose and delete the composite object.
                # Need to create a unique composite name to avoid mutation on the same object among processes.
                composed_object_name = COMPOSED_PREFIX + str(uuid.uuid4())
                composed_object = compose(
                    project_name,
                    bucket_name,
                    composed_object_name,
                    objects_slice,
                    storage_client,
                    retry_config=retry_config,
                )
                current_composed_object = composed_object
                yield from (decompose(
                    project_name,
                    bucket_name,
                    composed_object_name,
                    objects_slice,
                    storage_client,
                    retry_config=retry_config,
                ))

                try:
                    composed_object.delete(retry=retry_config)
                    current_composed_object = None
                except Exception as e:
                    logging.exception(
                        f"exception while deleting the composite object: {e}")


def clean_composed_object(composed_object):
    if composed_object:
        try:
            composed_object.delete(retry=MODIFIED_RETRY)
        except Exception as e:
            logging.exception(
                f"exception while deleting composite object: {e}")


def term_signal_handler(signal_num, frame):
    print("Ctrl+C interrupt detected. Cleaning up and exiting...")
    clean_composed_object(current_composed_object)
    sys.exit(0)



"""
 Copyright 2023 Google LLC

 Licensed under the Apache License, Version 2.0 (the "License");
 you may not use this file except in compliance with the License.
 You may obtain a copy of the License at

      https://www.apache.org/licenses/LICENSE-2.0

 Unless required by applicable law or agreed to in writing, software
 distributed under the License is distributed on an "AS IS" BASIS,
 WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 See the License for the specific language governing permissions and
 limitations under the License.
 """

from __future__ import annotations

from collections.abc import Sequence
from dataclasses import dataclass
from fractions import Fraction
from itertools import count


@dataclass
class MinimalIntRange:
    start_int: int
    end_int: int
    min_len: int


@dataclass
class GenerateSplitsOpts:
    min_int_range: MinimalIntRange
    num_splits: int
    start_range: str
    end_range: str


class RangeSplitter(object):
    """Manages splits performed to facilitate the work-stealing algorithm.

    Attr:
      alphabet_map: An int mapping for an alphabet of arbitrary character size.
      sorted_alphabet: The sorted alphabet that initializes the RangeSplitter.
    """

    min_splits = 2

    def __init__(self, alphabet_map: dict[int, str],
                 sorted_alphabet: Sequence[str]):
        self.alphabet_map = alphabet_map
        self.sorted_alphabet = sorted_alphabet
        self.alphabet_set = set(sorted_alphabet)

    def split_range(
        self,
        start_range: str,
        end_range: str,
        num_splits: int,
    ) -> Sequence[str]:
        """Creates a given number of splits based on a provided start and end range.

        Args:
          start_range (str): The string marking the start of the split range.
          end_range (str): The string marking the end of the split range.
          num_splits (int): The number of splitpoints to return.

        Returns:
          A sequence of split points dividing up the provided range.
        """
        if num_splits < 1:
            raise ValueError("Got num_splits of %s but need minimum of %s." %
                             (num_splits, self.min_splits))
        if len(end_range) != 0 and start_range >= end_range:
            return []

        if self.is_range_equal_with_padding(start_range, end_range):
            return []

        self.add_characters_to_alphabet(start_range + end_range)

        min_int_range = self.string_to_minimal_int_range(
            start_range, end_range, num_splits)

        split_points = self.generate_splits(
            GenerateSplitsOpts(min_int_range, num_splits, start_range,
                               end_range))
        return split_points

    def generate_splits(self, opts: GenerateSplitsOpts) -> Sequence[str]:
        """Generates a list of split points.

        Args:
          opts (GenerateSplitOpts): Set of options for generating splitpoints

        Returns:
          A list of split points.
        """
        start_int = opts.min_int_range.start_int
        end_int = opts.min_int_range.end_int
        min_len = opts.min_int_range.min_len

        range_diff = end_int - start_int
        split_points = []
        range_interval = opts.num_splits + 1
        adjustment = Fraction(range_diff / range_interval)

        for i in range(1, opts.num_splits + 1):
            split_point = start_int + adjustment * i
            split_string = self.int_to_string(int(split_point), min_len)

            is_greater_than_start = (len(split_string) > 0
                                     and split_string > opts.start_range)
            is_less_than_end = len(
                opts.end_range) == 0 or (len(split_string) > 0
                                         and split_string < opts.end_range)

            if is_greater_than_start and is_less_than_end:
                split_points.append(split_string)

        return split_points

    def int_to_string(self, split_point: int, string_len: int) -> str:
        """Converts the base len(alphabet) int back into a string.

        Args:
          split_point (int): A valid split point int to be converted to string.
          string_len (int): The required length of the resulting string.

        Returns:
          A string derived from a base len(alphabet) int.
        """
        alphabet_len = len(self.sorted_alphabet)
        split_string = ""

        for _ in range(string_len):
            remainder = split_point % alphabet_len
            split_point //= alphabet_len
            split_string += self.sorted_alphabet[remainder]

        # This is assembeled backwards via division, so we reverse the final string.
        return split_string[::-1]

    def string_to_minimal_int_range(self, start_range: str, end_range: str,
                                    num_splits: int) -> MinimalIntRange:
        """Converts a string range to a minimal integer range.

        Args:
          start_range (str): The string marking the start of the split range.
          end_range (str): The string marking the end of the split range.
          num_splits (int): The number of splitpoints to return.

        Returns:
          A minimal integer range.
        """

        start_int = 0
        end_int = 0

        alphabet_len = len(self.sorted_alphabet)
        start_char = self.sorted_alphabet[0]
        end_char = self.sorted_alphabet[-1]

        end_default_char = start_char
        if len(end_range) == 0:
            end_default_char = end_char

        for i in count(0):
            start_pos = self.alphabet_map[get_char_or_default(
                start_range, i, start_char)]
            start_int *= alphabet_len
            start_int += start_pos

            end_pos = self.alphabet_map[get_char_or_default(
                end_range, i, end_default_char)]
            end_int *= alphabet_len
            end_int += end_pos

            difference = end_int - start_int
            if difference > num_splits:
                # Due to zero indexing, min length must have 1 added to it.
                return MinimalIntRange(start_int, end_int, i + 1)

    def is_range_equal_with_padding(self, start_range: str, end_range: str):
        """Checks for equality between two string ranges.

        Args:
          start_range (str): The start range for the split.
          end_range (str): The end range for the split.

        Returns:
          Boolean indicating equality of the two provided ranges.
        """

        if len(end_range) == 0:
            return False

        longest = max(len(start_range), len(end_range))

        smallest_char = self.sorted_alphabet[0]

        for i in range(longest):
            char_start = get_char_or_default(start_range, i, smallest_char)
            char_end = get_char_or_default(end_range, i, smallest_char)

            if char_start != char_end:
                return False

        return True

    def add_characters_to_alphabet(self, characters: str):
        """Adds a character to the known alphabet.

        Args:
          characters: The string of characters to add to the library.
        """
        unique_characters = set(characters)
        new_alphabet = self.alphabet_set.union(unique_characters)
        if len(new_alphabet) != len(self.alphabet_set):
            self.sorted_alphabet = sorted(new_alphabet)
            self.alphabet_map = {
                val: index
                for index, val in enumerate(self.sorted_alphabet)
            }


def get_char_or_default(characters: str, index: int, default_char: str) -> str:
    """Returns the character at the given index or the default character if the index is out of bounds.

    Args:
      characters (str): The range string to check.
      index (int): The current iteration index across characters.
      default_char (str): The smallest character in the implemented char set.

    Returns:
      The resulting character for the given index.
    """
    if index < 0 or index >= len(characters):
        return default_char

    return characters[index]


def new_rangesplitter(alphabet: str) -> RangeSplitter:
    """Creates a new RangeSplitter with the given alphabets.

    Note that the alphabets are a predetermined set of characters
    by the work-stealing algorithm, and the characters are guaranteed to be unique.

    Args:
      alphabet (str): The full set of characters used for this range splitter.

    Returns:
      An instance of the RangeSplitter class that is used to manage splits
      performed to facilitate the work-stealing algorithm.
    """
    if len(alphabet) == 0:
        raise ValueError("Cannot split with an empty alphabet.")
    sorted_alphabet = sorted(alphabet)
    alphabet_map = {val: index for index, val in enumerate(sorted_alphabet)}
    return RangeSplitter(alphabet_map, sorted_alphabet)
